"""
Copyright 2021 Objectiv B.V.
"""

# Any import from modelhub initializes all the types, do not remove
from modelhub import __version__
import pytest
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests.functional.bach.test_data_and_utils import assert_equals_data
from uuid import UUID

# map
def test_is_first_session(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')

    s = modelhub.map.is_first_session(df)

    assert_equals_data(
        s,
        expected_columns=['event_id', 'is_first_session'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # use created series to filter dataframe
    df['s'] = s

    assert_equals_data(
        df[df.s].s,
        expected_columns=['event_id', 's'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )


def test_is_new_user(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')

    s = modelhub.map.is_new_user(df)

    assert_equals_data(
        s,
        expected_columns=['event_id', 'is_new_user'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # use created series to filter dataframe
    df['s'] = s

    assert_equals_data(
        df[df.s].s,
        expected_columns=['event_id', 's'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    s = modelhub.map.is_new_user(df, time_aggregation='%Y-%m')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'is_new_user'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )


def test_add_conversion_event(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')

    location_stack = df.location_stack.json[{'_type': 'LinkContext', 'id': 'cta-repo-button'}:]
    event_type = 'ClickEvent'
    conversion = 'github_clicks'
    modelhub.add_conversion_event(location_stack=location_stack,
                                  event_type=event_type,
                                  name=conversion)

    assert isinstance(modelhub._conversion_events, dict)
    assert len(modelhub._conversion_events) == 1
    assert modelhub._conversion_events[conversion] == (location_stack, event_type)

    ser = modelhub.map.is_conversion_event(df, 'github_clicks')
    assert_equals_data(
        ser,
        expected_columns=['event_id', 'is_conversion_event'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), False]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # location_stack not set
    modelhub.add_conversion_event(event_type=event_type, name=conversion)
    assert modelhub._conversion_events[conversion] == (None, event_type)

    ser = modelhub.map.is_conversion_event(df, 'github_clicks')

    assert_equals_data(
        ser,
        expected_columns=['event_id', 'is_conversion_event'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # event_type not set
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    modelhub.add_conversion_event(location_stack=location_stack, name='github_clicks')
    assert len(modelhub._conversion_events) == 1
    assert modelhub._conversion_events[conversion] == (location_stack, None)

    ser = modelhub.map.is_conversion_event(df, 'github_clicks')

    assert_equals_data(
        ser,
        expected_columns=['event_id', 'is_conversion_event'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), False]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # name not set
    modelhub.add_conversion_event(event_type='ClickEvent')
    assert len(modelhub._conversion_events) == 2
    assert modelhub._conversion_events['conversion_2'] == (None, event_type)

    ser = modelhub.map.is_conversion_event(df, 'conversion_2')

    assert_equals_data(
        ser,
        expected_columns=['event_id', 'is_conversion_event'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )


def test_is_conversion_event(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')

    # add conversion event
    modelhub.add_conversion_event(location_stack=df.location_stack.json[{'_type': 'LinkContext',
                                                                         'id': 'cta-repo-button'}:],
                                  event_type='ClickEvent', name='github_clicks')
    s = modelhub.map.is_conversion_event(df, 'github_clicks')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'is_conversion_event'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), False],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), False]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # use created series to filter dataframe
    df['s'] = s

    assert_equals_data(
        df[df.s].s,
        expected_columns=['event_id', 's'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # wrong conversion_event name
    with pytest.raises(KeyError, match="not labeled as a conversion"):
        modelhub.map.is_conversion_event(df, 'some_clicks')

    with pytest.raises(KeyError, match="not labeled as a conversion"):
        modelhub.map.is_conversion_event(df, None)


def test_conversions_counter(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')

    # add conversion event
    modelhub.add_conversion_event(location_stack=df.location_stack.json[{'_type': 'LinkContext', 'id': 'cta-repo-button'}:],
                            event_type='ClickEvent',
                            name='github_clicks')
    s = modelhub.map.conversions_counter(df, 'github_clicks')

    assert_equals_data(
        s,

        expected_columns=['event_id', 'converted'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), 0]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # use created series to filter dataframe
    df['s'] = s == 1

    assert_equals_data(
        df[df.s].s,
        expected_columns=['event_id', 's'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), 1]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    s = modelhub.map.conversions_counter(df, 'github_clicks', partition='user_id')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'converted'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), 1]
        ],
        order_by='event_id',
        convert_uuid=True,
    )


def test_conversions_in_time(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')

    # add conversion event
    modelhub.add_conversion_event(
        location_stack=df.location_stack.json[{'_type': 'LinkContext', 'id': 'cta-repo-button'}:],
        event_type='ClickEvent',
        name='github_clicks')
    s = modelhub.map.conversions_in_time(df, 'github_clicks')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'conversions_in_time'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), 0],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), 0]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # use created series to filter dataframe
    df['s'] = s == 1

    assert_equals_data(
        df[df.s].s,
        expected_columns=['event_id', 's'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), True]
        ],
        order_by='event_id',
        convert_uuid=True,
    )


def test_pre_conversion_hit_number(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')

    # add conversion event
    modelhub.add_conversion_event(location_stack=df.location_stack.json[{'_type': 'LinkContext', 'id': 'cta-repo-button'}:],
                            event_type='ClickEvent',
                            name='github_clicks')
    s = modelhub.map.pre_conversion_hit_number(df, 'github_clicks')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'pre_conversion_hit_number'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), 2],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), None]
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    # use created series to filter dataframe
    df['s'] = s == 2

    assert_equals_data(
        df[df.s].s,
        expected_columns=['event_id', 's'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), True],
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    s = modelhub.map.pre_conversion_hit_number(df, 'github_clicks', partition='user_id')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'pre_conversion_hit_number'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), 2],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), 1],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), None],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), 4],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), 3]
        ],
        order_by='event_id',
        convert_uuid=True,
    )


def test_time_agg(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params)
    s = modelhub.time_agg(df)

    assert_equals_data(
        s,
        expected_columns=['event_id', 'time_aggregation'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), '2021-11-30 10:23:36.287000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), '2021-11-30 10:23:36.290000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), '2021-11-30 10:23:36.291000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), '2021-11-30 10:23:36.267000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), '2021-12-01 10:23:36.276000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), '2021-12-01 10:23:36.279000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), '2021-12-02 10:23:36.281000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), '2021-12-02 10:23:36.281000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), '2021-12-02 14:23:36.282000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), '2021-12-03 10:23:36.283000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), '2021-11-29 10:23:36.286000'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), '2021-11-29 10:23:36.287000']
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m')
    s = modelhub.time_agg(df)

    assert_equals_data(
        s,
        expected_columns=['event_id', 'time_aggregation'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), '2021-11'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), '2021-11'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), '2021-11'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), '2021-11'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), '2021-12'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), '2021-12'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), '2021-12'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), '2021-12'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), '2021-12'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), '2021-12'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), '2021-11'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), '2021-11']
        ],
        order_by='event_id',
        convert_uuid=True,
    )

    df, modelhub = get_objectiv_dataframe_test(db_params)
    s = modelhub.time_agg(df, time_aggregation='%Y-%m-%d')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'time_aggregation'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), '2021-11-30'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), '2021-11-30'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), '2021-11-30'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), '2021-11-30'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), '2021-12-01'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), '2021-12-01'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), '2021-12-02'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), '2021-12-02'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), '2021-12-02'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), '2021-12-03'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), '2021-11-29'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), '2021-11-29']],
        order_by='event_id',
        convert_uuid=True,
    )

    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.time_agg(df, time_aggregation='%Y')

    assert_equals_data(
        s,
        expected_columns=['event_id', 'time_aggregation'],
        expected_data=[
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac301'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac302'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac303'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac304'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac305'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac306'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac307'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac308'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac309'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac310'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac311'), '2021'],
            [UUID('12b55ed5-4295-4fc1-bf1f-88d64d1ac312'), '2021']],
        order_by='event_id',
        convert_uuid=True,
    )



"""
Copyright 2021 Objectiv B.V.
"""

# Any import from modelhub initializes all the types, do not remove
from modelhub import __version__
import pytest
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests.functional.bach.test_data_and_utils import assert_equals_data
import datetime


def test_defaults(db_params):
    # setting nothing, thus using all defaults (which is just moment without formatting)
    df, modelhub = get_objectiv_dataframe_test(db_params)
    s = modelhub.aggregate.session_duration(df)

    # with standard time_aggregation, all sessions are bounces
    assert len(s.to_numpy()) == 0


@pytest.mark.parametrize("exclude_bounces,expected_data", [
    (True, [[datetime.timedelta(microseconds=2667)]]),
    (False, [[datetime.timedelta(microseconds=1143)]])
])
def test_no_grouping(db_params, exclude_bounces, expected_data):
    # not grouping to anything
    df, modelhub = get_objectiv_dataframe_test(db_params)
    s = modelhub.aggregate.session_duration(df, groupby=None, exclude_bounces=exclude_bounces)

    assert_equals_data(
        s,
        expected_columns=['session_duration'],
        expected_data=expected_data,
        use_to_pandas=True,
    )


def test_time_aggregation_in_df(db_params):
    # using time_aggregation (and default groupby: mh.time_agg(df, ))
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.aggregate.session_duration(df)

    assert_equals_data(
        s,
        expected_columns=['time_aggregation', 'session_duration'],
        expected_data=[
            ['2021-11-29', datetime.timedelta(microseconds=1000)],
            ['2021-11-30', datetime.timedelta(microseconds=4000)],
            ['2021-12-01', datetime.timedelta(microseconds=3000)]
        ],
        use_to_pandas=True,
    )


def test_overriding_time_aggregation_in(db_params):
    # overriding time_aggregation
    df, modelhub = get_objectiv_dataframe_test(db_params)
    bts = modelhub.aggregate.session_duration(df, groupby=modelhub.time_agg(df, '%Y-%m'))

    assert_equals_data(
        bts,
        expected_columns=['time_aggregation', 'session_duration'],
        expected_data=[
            ['2021-11', datetime.timedelta(microseconds=2500)],
            ['2021-12', datetime.timedelta(microseconds=3000)]
        ],
        use_to_pandas=True,
    )

    bts = modelhub.aggregate.session_duration(df, groupby=modelhub.time_agg(df, '%Y-%m'), method='sum')

    assert_equals_data(
        bts,
        expected_columns=['time_aggregation', 'session_duration'],
        expected_data=[
            ['2021-11', datetime.timedelta(microseconds=5000)],
            ['2021-12', datetime.timedelta(microseconds=3000)],
        ],
        use_to_pandas=True,
    )


def test_groupby(db_params):
    # group by other columns
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.aggregate.session_duration(df, groupby='event_type')

    assert_equals_data(
        s,
        expected_columns=['event_type', 'session_duration'],
        expected_data=[['ClickEvent', datetime.timedelta(microseconds=2667)]],
        use_to_pandas=True,
    )


def test_groupby_incl_time_agg(db_params):
    # group by other columns (as series), including time_agg
    df, modelhub = get_objectiv_dataframe_test(db_params, time_aggregation='%Y-%m-%d')
    s = modelhub.aggregate.session_duration(df, groupby=[modelhub.time_agg(df, '%Y-%m'), df.session_id])

    assert_equals_data(
        s,
        expected_columns=['time_aggregation', 'session_id', 'session_duration'],
        expected_data=[
            ['2021-11', 1, datetime.timedelta(microseconds=1000)],
            ['2021-11', 3, datetime.timedelta(microseconds=4000)],
            ['2021-12', 4, datetime.timedelta(microseconds=3000)]
        ],
        use_to_pandas=True,
    )



"""
Copyright 2022 Objectiv B.V.
"""


# Any import from modelhub initializes all the types, do not remove
from modelhub import __version__
import pytest
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests.functional.bach.test_data_and_utils import assert_equals_data


def test_retention_matrix(db_params):

    df, modelhub = get_objectiv_dataframe_test(db_params)
    event_type = 'ClickEvent'

    # yearly
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='yearly',
                                               event_type=event_type,
                                               percentage=False,
                                               display=False)

    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0'],
        expected_data=[
            ['2021', 4],
        ],
        use_to_pandas=True,
    )

    # monthly
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='monthly',
                                               event_type=event_type,
                                               percentage=False,
                                               display=False)

    # filling nan values with -999 in order to be able to do the check
    # (nan values are causing a trouble)
    data = data.fillna(value=-999)
    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0', '_1'],
        expected_data=[
            ['2021-11', 2, 1],
            ['2021-12', 2, -999],
        ],
        use_to_pandas=True,
    )

    # weekly
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='weekly',
                                               event_type=event_type,
                                               percentage=False,
                                               display=False)
    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0'],
        expected_data=[
            ['2021-11-29', 4],
        ],
        use_to_pandas=True,
    )

    # daily
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='daily',
                                               event_type=event_type,
                                               percentage=False,
                                               display=False)

    data = data.fillna(value=-999)
    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0', '_1'],
        expected_data=[
            ['2021-11-29', 1, 1],
            ['2021-11-30', 1, 1],
            ['2021-12-02', 1, -999],
            ['2021-12-03', 1, -999],
        ],
        use_to_pandas=True,
    )

    # not supported time_period
    with pytest.raises(ValueError, match='biweekly time_period is not available.'):
        modelhub.aggregate.retention_matrix(df,
                                            event_type=event_type,
                                            time_period='biweekly',
                                            display=False)

    # non-existing event type
    data = modelhub.aggregate.retention_matrix(df,
                                               event_type='some_event',
                                               display=False)

    assert list(data.index.keys()) == ['first_cohort']
    assert data.columns == []

    # all events
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='yearly',
                                               event_type=None,
                                               percentage=False,
                                               display=False)

    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0'],
        expected_data=[
            ['2021', 4],
        ],
        use_to_pandas=True,
    )

    # percentage
    data = modelhub.aggregate.retention_matrix(df,
                                              time_period='monthly',
                                              event_type=event_type,
                                              percentage=True,
                                              display=False)

    data = data.fillna(value=-999.0)
    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0', '_1'],
        expected_data=[
            ['2021-11', 100.0, 50.0],
            ['2021-12', 100.0, -999.0],
        ],
        use_to_pandas=True,
    )

    # start_date
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='daily',
                                               event_type=event_type,
                                               start_date='2021-11-30',
                                               percentage=False,
                                               display=False)

    data = data.fillna(value=-999)
    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0', '_1'],
        expected_data=[
            ['2021-11-30', 1, 1],
            ['2021-12-02', 1, -999],
            ['2021-12-03', 1, -999],
        ],
        use_to_pandas=True,
    )

    # end_date
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='daily',
                                               event_type=event_type,
                                               end_date='2021-12-02',
                                               percentage=False,
                                               display=False)

    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0', '_1'],
        expected_data=[
            ['2021-11-29', 1, 1],
            ['2021-11-30', 1, 1],
        ],
        use_to_pandas=True,
    )

    # start_date and end_date
    data = modelhub.aggregate.retention_matrix(df,
                                               time_period='daily',
                                               event_type=event_type,
                                               start_date='2021-11-30',
                                               end_date='2021-12-02',
                                               percentage=False,
                                               display=False)

    assert_equals_data(
        data,
        expected_columns=['first_cohort', '_0', '_1'],
        expected_data=[
            ['2021-11-30', 1, 1],
        ],
        use_to_pandas=True,
    )

    # wrong start_date
    with pytest.raises(ValueError, match="time data '2021-11' does not match format '%Y-%m-%d"):
        modelhub.aggregate.retention_matrix(df,
                                            time_period='daily',
                                            event_type=event_type,
                                            start_date='2021-11',
                                            percentage=False,
                                            display=False)

    # wrong end_date
    with pytest.raises(ValueError, match="time data '2021-11' does not match format '%Y-%m-%d"):
        modelhub.aggregate.retention_matrix(df,
                                            time_period='daily',
                                            event_type=event_type,
                                            end_date='2021-11',
                                            percentage=False,
                                            display=False)



# all data below is generated dummy data
TEST_DATA_JSON_REAL = [
    [1,
     '[{"_type": "ApplicationContext", "id": "rod-web-demo"}, {"id": "http_context", "referrer": "https://rick.objectiv.io/", "user_agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0", "remote_address": "144.144.144.144", "_type": "HttpContext"}, {"id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "cookie_id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "_type": "CookieIdContext"}]',
     '[{"_type": "WebDocumentContext", "id": "#document", "url": "https://rick.objectiv.io/"}, {"_type": "SectionContext", "id": "home"}, {"_type": "SectionContext", "id": "yep"}, {"_type": "SectionContext", "id": "cc91EfoBh8A"}]'],
    [2,
     '[{"_type": "ApplicationContext", "id": "rod-web-demo"}, {"id": "http_context", "referrer": "https://rick.objectiv.io/", "user_agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0", "remote_address": "144.144.144.144", "_type": "HttpContext"}, {"id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "cookie_id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "_type": "CookieIdContext"}]',
     '[{"_type": "WebDocumentContext", "id": "#document", "url": "https://rick.objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "NavigationContext", "id": "navigation", "_types": ["AbstractContext", "AbstractLocationContext", "NavigationContext", "SectionContext"]}]'],
    [3,
     '[{"_type": "ApplicationContext", "id": "rod-web-demo"}, {"id": "http_context", "referrer": "https://rick.objectiv.io/", "user_agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0", "remote_address": "144.144.144.144", "_type": "HttpContext"}, {"id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "cookie_id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "_type": "CookieIdContext"}]',
     '[{"_type": "WebDocumentContext", "id": "#document", "url": "https://rick.objectiv.io/"}, {"_type": "SectionContext", "id": "home"}, {"_type": "SectionContext", "id": "new"}, {"_type": "SectionContext", "id": "BeyEGebJ1l4"}]'],
    [4,
     '[{"_type": "ApplicationContext", "id": "rod-web-demo"}, {"id": "http_context", "referrer": "https://rick.objectiv.io/", "user_agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0", "remote_address": "144.144.144.144", "_type": "HttpContext"}, {"id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "cookie_id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "_type": "CookieIdContext"}]',
     '[{"_type": "WebDocumentContext", "id": "#document", "url": "https://rick.objectiv.io/"}, {"_type": "SectionContext", "id": "home"}, {"_type": "SectionContext", "id": "new"}, {"_type": "SectionContext", "id": "yBwD4iYcWC4"}]'],
    [5,
     '[{"_type": "ApplicationContext", "id": "rod-web-demo"}, {"id": "http_context", "referrer": "https://rick.objectiv.io/", "user_agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0", "remote_address": "144.144.144.144", "_type": "HttpContext"}, {"id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "cookie_id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "_type": "CookieIdContext"}]',
     '[{"_type": "WebDocumentContext", "id": "#document", "url": "https://rick.objectiv.io/"}, {"_type": "SectionContext", "id": "home"}, {"_type": "SectionContext", "id": "new"}, {"_type": "MediaPlayerContext", "id": "eYuUAGXN0KM"}]'],
    [6,
     '[{"_type": "ApplicationContext", "id": "rod-web-demo"}, {"id": "http_context", "referrer": "https://rick.objectiv.io/", "user_agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:94.0) Gecko/20100101 Firefox/94.0", "remote_address": "144.144.144.144", "_type": "HttpContext"}, {"id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "cookie_id": "f84446c6-eb76-4458-8ef4-93ade596fd5b", "_type": "CookieIdContext"}]',
     '[{"_type": "WebDocumentContext", "id": "#document", "url": "https://rick.objectiv.io/"}]']

]
JSON_COLUMNS_REAL = ['event_id', 'global_contexts', 'location_stack']



"""
Copyright 2022 Objectiv B.V.
### Fixtures
There is some pytest 'magic' here that automatically fills out the db parameters for
test functions that require an engine.
By default such a test function will get a Postgres db parameters. But if --big-query or --all is
specified on the commandline, then it will (also) get a BigQuery db parameters. For specific
tests, it is possible to disable postgres or bigquery testing, see 'marks' section below.
### Marks and Test Categorization
A lot of functionality needs to be tested for multiple databases. The 'engine' and 'dialects' fixtures
mentioned above help with that. Additionally we have some marks (`@pytest.mark.<type>`) to make it explicit
which databases we expect tests to run against.
We broadly want 5 categories of tests:
* unit-test: These don't interact with a database
  * unit-tests that are tested with multiple database dialects (1)
  * unit-tests that are database-dialect independent (2)
* functional-tests: These interact with a database
  *  functional-tests that run against all supported databases (3)
  *  functional-tests that run against all supported databases except Postgres (4)
  *  functional-tests that run against all supported databases except BigQuery (5)
1 and 3 are the default for tests. These either get 'db_params' as fixture and run against all
databases. Category 2 are tests that test generic code that is not geared to a specific database.
Category 4 and 5 are for functionality that we explicitly not support on some databases.
Category 4, and 5 are the exception, these need to be marked with the `skip_postgres` or `skip_bigquery` marks.
"""
import os

from _pytest.python import Metafunc
from _pytest.config.argparsing import Parser
from tests_modelhub.data_and_utils.utils import DBParams


DB_PG_TEST_URL = os.environ.get('OBJ_DB_PG_TEST_URL', 'postgresql://objectiv:@localhost:5432/objectiv')
DB_BQ_TEST_URL = os.environ.get('OBJ_DB_BQ_TEST_URL', 'bigquery://objectiv-snowplow-test-2/modelhub_test')
DB_BQ_CREDENTIALS_PATH = os.environ.get(
    'OBJ_DB_BQ_CREDENTIALS_PATH',
    os.path.dirname(os.path.dirname(os.path.realpath(__file__))) + '/.secrets/bach-big-query-testing.json'
)


MARK_SKIP_POSTGRES = 'skip_postgres'
MARK_SKIP_BIGQUERY = 'skip_bigquery'


def pytest_addoption(parser: Parser):
    # Add options for parameterizing multi-database tests for testing either Postgres, Bigquery, or both.
    # The actual parameterizing happens in pytest_generate_tests(), based on the paramters that the user
    # provides

    # This function will automatically be called by pytest at the start of a test run, see:
    # https://docs.pytest.org/en/6.2.x/reference.html#initialization-hooks
    parser.addoption('--postgres', action='store_true', help='run the functional tests for Postgres')
    parser.addoption('--big-query', action='store_true', help='run the functional tests for BigQuery')
    parser.addoption('--all', action='store_true', help='run the functional tests for all databases.')


def pytest_generate_tests(metafunc: Metafunc):
    # Paramaterize the 'db_params' parameters of tests based on the options specified by the
    # user (see pytest_addoption() for options).

    # This function will automatically be called by pytest while it is creating the list of tests to run,
    # see: https://docs.pytest.org/en/6.2.x/reference.html#collection-hooks
    markers = list(metafunc.definition.iter_markers())
    skip_postgres = any(mark.name == MARK_SKIP_POSTGRES for mark in markers)
    skip_bigquery = any(mark.name == MARK_SKIP_BIGQUERY for mark in markers)
    db_params = []

    testing_pg = not metafunc.config.getoption("big_query")
    testing_bq = metafunc.config.getoption("all") or metafunc.config.getoption("big_query")

    if testing_pg and not skip_postgres:
        db_params.append(get_postgres_db_params())

    if testing_bq and not skip_bigquery:
        db_params.append(_get_bigquery_db_params())

    if 'db_params' in metafunc.fixturenames:
        metafunc.parametrize("db_params", db_params)


def get_postgres_db_params() -> DBParams:
    """
    Get Postgres DBParams. Never call this function from a test, always use the 'db_params' fixture.
    """
    return DBParams(
        url=DB_PG_TEST_URL,
        credentials=None,
        table_name='objectiv_data',
    )


def _get_bigquery_db_params() -> DBParams:
    return DBParams(
        url=DB_BQ_TEST_URL,
        credentials=DB_BQ_CREDENTIALS_PATH,
        table_name='events',
    )



"""
Copyright 2022 Objectiv B.V.
"""

# Any import from modelhub initializes all the types, do not remove
from modelhub import __version__
import pytest
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests.functional.bach.test_data_and_utils import assert_equals_data


def test_top_product_features_before_conversion(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params)
    initial_columns = df.data_columns

    event_type = 'ClickEvent'
    name = 'clicks'

    # name checks
    modelhub.add_conversion_event(name=name,
                                  event_type=event_type)
    with pytest.raises(ValueError, match='Conversion event label is not provided.'):
        modelhub.aggregate.top_product_features_before_conversion(df, name=None)

    with pytest.raises(KeyError, match='Key some_name is not labeled as a conversion'):
        modelhub.aggregate.top_product_features_before_conversion(df, name='some_name')

    # without location_stack
    modelhub.add_conversion_event(event_type=event_type, name=name)
    cdf = modelhub.aggregate.top_product_features_before_conversion(df, name=name)

    # index checks
    expected_index = ['application', 'feature_nice_name', 'event_type']
    assert len(cdf.index) == 3
    for _index in expected_index:
        assert _index in cdf.index

    # data checks
    assert cdf.data_columns == ['unique_users']
    assert list(cdf.data.values()) == [1]

    # with location_stack
    location_stack = df.location_stack.json[{'id': 'main'}:]
    modelhub.add_conversion_event(location_stack=location_stack,
                                  event_type=event_type,
                                  name=name)
    cdf = modelhub.aggregate.top_product_features_before_conversion(df, name=name)
    assert len(cdf.index) == 3

    feature_name = 'Link: GitHub located at Web Document: #document => Section:' \
                   ' navbar-top => Overlay: hamburger-menu'
    assert_equals_data(
        cdf,
        expected_columns=['application', 'feature_nice_name', 'event_type', 'unique_users'],
        expected_data=[
            ['objectiv-website', feature_name, 'ClickEvent', 1],
        ],
        use_to_pandas=True
    )

    # check if any new column is added to the original dataframe
    assert sorted(initial_columns) == sorted(df.data_columns)




"""
Copyright 2021 Objectiv B.V.
"""
from modelhub.pipelines.extracted_contexts import ExtractedContextsPipeline
from modelhub.pipelines.sessionized_data import SessionizedDataPipeline
from modelhub.pipelines.identity_resolution import IdentityResolutionPipeline



import json
from typing import List, Dict

_BASE_TEST_DATA_OBJECTIVE = [
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac301','2021-11-30','2021-11-30 10:23:36.287','b2df75d2-d7ca-48ac-9747-af47d7a4a2b2','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "navbar-top", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "OverlayContext", "id": "hamburger-menu", "_types": ["AbstractContext", "AbstractLocationContext", "OverlayContext", "SectionContext"]}, {"_type": "LinkContext", "id": "GitHub", "text": "GitHub", "href": "https://github.com/objectiv", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "cookie_id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "729c84f9-91d0-4f9f-be58-5cfb2d8130e4", "time": 1636476263115, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac302','2021-11-30','2021-11-30 10:23:36.290','b2df75d2-d7ca-48ac-9747-af47d7a4a2b2','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "main", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "SectionContext", "id": "location-stack", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "LinkContext", "id": "cta-docs-location-stack", "text": "Docs - Location Stack", "href": "/docs/taxonomy", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "cookie_id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "1049e11c-bb9c-4b84-9dac-b4125998999d", "time": 1636475896879, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac303','2021-11-30','2021-11-30 10:23:36.291','b2df75d2-d7ca-48ac-9747-af47d7a4a2b2','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "header", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "LinkContext", "id": "cta-repo-button", "text": "Objectiv on GitHub", "href": "https://github.com/objectiv/objectiv-analytics", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "cookie_id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "fd8239de-032f-499a-9849-8e97214ecdf1", "time": 1636475880112, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac304','2021-11-30','2021-11-30 10:23:36.267','b2df75d2-d7ca-48ac-9747-af47d7a4a2b1','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/docs/modeling/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "LinkContext", "id": "notebook-product-analytics", "text": "sandboxed notebook", "href": "https://notebook.objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-docs", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "cookie_id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "a789d8fe-5cd9-4ff0-9780-a56cf094b62a", "time": 1636475922156, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac305','2021-12-01','2021-12-01 10:23:36.276','b2df75d2-d7ca-48ac-9747-af47d7a4a2b1','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "navbar-top", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "OverlayContext", "id": "hamburger-menu", "_types": ["AbstractContext", "AbstractLocationContext", "OverlayContext", "SectionContext"]}, {"_type": "LinkContext", "id": "About Us", "text": "About Us", "href": "about", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "cookie_id": "a30c5ca2-6f0c-4e56-997c-2148bd71ee8d", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "67cbfc73-b8bd-40f6-aa8e-88cb73857d09", "time": 1636475947689, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac306','2021-12-01','2021-12-01 10:23:36.279','b2df75d2-d7ca-48ac-9747-af47d7a4a2b1','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://www.objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "navbar-top", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "OverlayContext", "id": "hamburger-menu", "_types": ["AbstractContext", "AbstractLocationContext", "OverlayContext", "SectionContext"]}, {"_type": "LinkContext", "id": "Contact Us", "text": "Contact Us", "href": "mailto:hi@objectiv.io", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://www.objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "cookie_id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "899c18aa-a908-43f9-9827-d4b9072205ea", "time": 1636475983057, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac307','2021-12-02','2021-12-02 10:23:36.281','b2df75d2-d7ca-48ac-9747-af47d7a4a2b3','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://www.objectiv.io/jobs", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "footer", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "LinkContext", "id": "Cookies", "text": "Cookies", "href": "/privacy/cookies", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://www.objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "cookie_id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "837ae9db-497c-4925-a4c9-b2183bd3056b", "time": 1636476007981, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac308','2021-12-02','2021-12-02 10:23:36.281','b2df75d2-d7ca-48ac-9747-af47d7a4a2b3','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/docs/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "navbar-top", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "OverlayContext", "id": "hamburger-menu", "_types": ["AbstractContext", "AbstractLocationContext", "OverlayContext", "SectionContext"]}, {"_type": "ExpandableSectionContext", "id": "The Project", "_types": ["AbstractContext", "AbstractLocationContext", "ExpandableSectionContext", "SectionContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-docs", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 12; Pixel 4a) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "cookie_id": "1cc3cb08-010b-465a-8241-88c9b4d233ea", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "5835d00e-4099-44cc-9191-8baccc2d32fa", "time": 1636476074003, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac309','2021-12-02','2021-12-02 14:23:36.282','b2df75d2-d7ca-48ac-9747-af47d7a4a2b3','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "navbar-top", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "LinkContext", "id": "About Us", "text": "About Us", "href": "about", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.81 Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "fbca9fc6-4b76-459e-968c-0ecf3c78de4d", "cookie_id": "fbca9fc6-4b76-459e-968c-0ecf3c78de4d", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "690ada97-c0fa-4378-9c04-bd1f7753505a", "time": 1636476111218, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac310','2021-12-03','2021-12-03 10:23:36.283','b2df75d2-d7ca-48ac-9747-af47d7a4a2b4','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/about", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "navbar-top", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "OverlayContext", "id": "hamburger-menu", "_types": ["AbstractContext", "AbstractLocationContext", "OverlayContext", "SectionContext"]}, {"_type": "LinkContext", "id": "Docs", "text": "Docs", "href": "https://objectiv.io/docs/", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 10; POCOPHONE F1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "5b1e395f-ef4c-438c-aab2-ae0aa19131ee", "cookie_id": "5b1e395f-ef4c-438c-aab2-ae0aa19131ee", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "089ff754-35d6-49da-bb32-dc9031b10289", "time": 1636476142139, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac311','2021-11-29','2021-11-29 10:23:36.286','b2df75d2-d7ca-48ac-9747-af47d7a4a2b2','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "main", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "SectionContext", "id": "taxonomy", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "LinkContext", "id": "cta-docs-taxonomy", "text": "Docs - Taxonomy", "href": "/docs/taxonomy/", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-website", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 11; SM-G986B) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "81a8ace2-273b-4b95-b6a6-0fba33858a22", "cookie_id": "81a8ace2-273b-4b95-b6a6-0fba33858a22", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "fd54aa9a-b8b8-4feb-968d-8fa9f736c596", "time": 1636476191693, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}'),
    ('12b55ed5-4295-4fc1-bf1f-88d64d1ac312','2021-11-29','2021-11-29 10:23:36.287','b2df75d2-d7ca-48ac-9747-af47d7a4a2b2','{"_type": "ClickEvent", "location_stack": [{"_type": "WebDocumentContext", "id": "#document", "url": "https://objectiv.io/docs/taxonomy/", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext", "WebDocumentContext"]}, {"_type": "SectionContext", "id": "navbar-top", "_types": ["AbstractContext", "AbstractLocationContext", "SectionContext"]}, {"_type": "LinkContext", "id": "logo", "text": "Objectiv Documentation Logo", "href": "/docs/", "_types": ["AbstractContext", "AbstractLocationContext", "ActionContext", "ItemContext", "LinkContext"]}], "global_contexts": [{"_type": "ApplicationContext", "id": "objectiv-docs", "_types": ["AbstractContext", "AbstractGlobalContext", "ApplicationContext"]}, {"id": "http_context", "referrer": "https://objectiv.io/", "user_agent": "Mozilla/5.0 (Linux; Android 11; SM-G986B) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.74 Mobile Safari/537.36", "_type": "HttpContext", "_types": ["AbstractContext", "AbstractGlobalContext", "HttpContext"]}, {"id": "81a8ace2-273b-4b95-b6a6-0fba33858a22", "cookie_id": "81a8ace2-273b-4b95-b6a6-0fba33858a22", "_type": "CookieIdContext", "_types": ["AbstractContext", "AbstractGlobalContext", "CookieIdContext"]}], "id": "e2445152-327a-466f-a2bf-116f0146ab7a", "time": 1636476196460, "_types": ["AbstractEvent", "ClickEvent", "InteractiveEvent"]}')
]

IDENTITY_CONTEXTS_PER_EVENT = [
    # event: 301, user: a2b2
    [
        {'_type': 'IdentityContext', 'id': 'email', 'value': 'fake1@objectiv.io'},
        {'_type': 'IdentityContext', 'id': 'phone', 'value': '123456789'},
    ],
    # event: 302, user: a2b2
    [],
    # event: 303, user: a2b2
    [
        {'_type': 'IdentityContext', 'id': 'email', 'value': 'fake2@objectiv.io'},
    ],
    # event: 304, user: a2b1
    [],
    # event: 305, user: a2b1
    [{'_type': 'IdentityContext', 'id': 'phone', 'value': '123456789'}],
    # event: 306, user: a2b1
    [],
    # event: 307, user: a2b3
    [
        {'_type': 'IdentityContext', 'id': 'email', 'value': 'fake3@objectiv.io'},
        {'_type': 'IdentityContext', 'id': 'email', 'value': 'fake31@objectiv.io'},
    ],
    # event: 308, user: a2b3
    [],
    # event: 309, user: a2b3
    [],
    # event: 310, user: a2b4
    [],
    # event: 311, user: a2b2
    [{'_type': 'IdentityContext', 'id': 'email', 'value': 'fake4@objectiv.io'}],
    # event: 312, user: a2b2
    [],
]


def _add_identity_contexts_to_value_json(
    value: str, identity_contexts: List[Dict[str, str]]
) -> str:
    value_json = json.loads(value)
    value_json['global_contexts'].extend(identity_contexts)
    return json.dumps(value_json)


TEST_DATA_OBJECTIV = [
    (
        base_data[0],  # event_id
        base_data[1],  # day
        base_data[2],  # moment
        base_data[3],  # cookie_id
        _add_identity_contexts_to_value_json(base_data[4], identity_data),  # value

    )
    for base_data, identity_data in zip(_BASE_TEST_DATA_OBJECTIVE, IDENTITY_CONTEXTS_PER_EVENT)
]



"""
Copyright 2021 Objectiv B.V.
"""
from enum import Enum

import bach
from bach.partitioning import WindowFrameBoundary, WindowFrameMode
from typing import TYPE_CHECKING, Dict, List


from modelhub.decorators import use_only_required_objectiv_series

if TYPE_CHECKING:
    from modelhub import ModelHub


class _CalculatedConversionSeries(str, Enum):
    IS_CONVERSION_EVENT = 'is_conversion_event'
    CONVERSIONS_IN_TIME = 'conversions_in_time'
    CONVERSIONS_COUNTER = 'converted'
    PRE_CONVERSION_HIT_NUMBER = 'pre_conversion_hit_number'

    @property
    def private_name(self) -> str:
        return f'__{self}'

    @property
    def public_name(self) -> str:
        return self.value


_DEPENDENCIES_PER_CONVERSION_SERIES: Dict[_CalculatedConversionSeries, List[_CalculatedConversionSeries]] = {
    _CalculatedConversionSeries.CONVERSIONS_IN_TIME: [
        _CalculatedConversionSeries.IS_CONVERSION_EVENT,
    ],
    _CalculatedConversionSeries.CONVERSIONS_COUNTER: [
        _CalculatedConversionSeries.IS_CONVERSION_EVENT,
        _CalculatedConversionSeries.CONVERSIONS_IN_TIME,
    ],
    _CalculatedConversionSeries.PRE_CONVERSION_HIT_NUMBER: [
        _CalculatedConversionSeries.IS_CONVERSION_EVENT,
        _CalculatedConversionSeries.CONVERSIONS_IN_TIME,
    ]
}


class Map:
    """
    Methods in this class can be used to map data in a DataFrame with Objectiv data to series values.

    Always returns Series with same index as the DataFrame the method is applied to, so the result can be set
    as columns to that DataFrame.
    """

    def __init__(self, mh: 'ModelHub'):
        self._mh = mh

    @use_only_required_objectiv_series(required_series=['user_id', 'session_id'])
    def is_first_session(self, data: bach.DataFrame) -> bach.SeriesBoolean:
        """
        Labels all hits in a session True if that session is the first session of that user in the data.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :returns: :py:class:`bach.SeriesBoolean` with the same index as ``data``.
        """

        window = data.groupby('user_id').window(
            mode=WindowFrameMode.ROWS,
            start_boundary=WindowFrameBoundary.PRECEDING,
            start_value=None,
            end_boundary=WindowFrameBoundary.FOLLOWING,
            end_value=None)

        first_session = window['session_id'].min()
        series = first_session == data.session_id

        new_series = series.copy_override(name='is_first_session',
                                          index=data.index).materialize()

        return new_series

    @use_only_required_objectiv_series(required_series=['user_id', 'session_id', 'moment'])
    def is_new_user(self, data: bach.DataFrame, time_aggregation: str = None) -> bach.SeriesBoolean:
        """
        Labels all hits True if the user is first seen in the period given `time_aggregation`.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param time_aggregation: if None, it uses the :py:attr:`ModelHub.time_aggregation` set in ModelHub
            instance.
        :returns: :py:class:`bach.SeriesBoolean` with the same index as ``data``.
        """

        frame_args = {
            'mode': WindowFrameMode.ROWS,
            'start_boundary': WindowFrameBoundary.PRECEDING,
            'end_boundary': WindowFrameBoundary.FOLLOWING,
        }
        data_cp = data.copy()
        data_cp['time_agg'] = self._mh.time_agg(data, time_aggregation)

        window = data_cp.groupby('user_id').window(**frame_args)
        window_ta = data_cp.groupby(['time_agg', 'user_id']).window(**frame_args)

        session_id_series = data_cp['session_id']
        is_first_session = session_id_series.min(partition=window)
        is_first_session_time_aggregation = session_id_series.min(partition=window_ta)

        is_new_user_series = is_first_session_time_aggregation == is_first_session
        is_new_user_series = is_new_user_series.copy_override_type(bach.SeriesBoolean)
        return is_new_user_series.copy_override(name='is_new_user').materialize()

    @use_only_required_objectiv_series(required_series=['event_type'])
    def is_conversion_event(self, data: bach.DataFrame, name: str) -> bach.SeriesBoolean:
        """
        Labels a hit True if it is a conversion event, all other hits are labeled False.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param name: the name of the conversion to label as set in
            :py:attr:`ModelHub.conversion_events`.
        :returns: :py:class:`bach.SeriesBoolean` with the same index as ``data``.
        """

        series_to_calculate = _CalculatedConversionSeries.IS_CONVERSION_EVENT
        conversion_df = self._get_calculated_conversion_df(
            series_to_calculate=series_to_calculate,
            data=data,
            name=name,
        )

        series = conversion_df[series_to_calculate.private_name]
        return (
            series
            .copy_override(name=series_to_calculate.public_name)
            .copy_override_type(bach.SeriesBoolean)
        )

    @use_only_required_objectiv_series(
        required_series=['session_id', 'moment', 'event_type'],
        include_series_from_params=['partition'],
    )
    def conversions_counter(self,
                            data: bach.DataFrame,
                            name: str,
                            partition: str = 'session_id') -> bach.SeriesInt64:
        """
        Counts the total number of conversions given a partition (ie session_id
        or user_id).

        :param name: the name of the conversion to label as set in
            :py:attr:`ModelHub.conversion_events`.
        :param partition: the partition over which the number of conversions are counted. Can be any column
            of the ObjectivFrame
        :returns: :py:class:`bach.SeriesBoolean` with the same index as ``data``.
        """

        series_to_calculate = _CalculatedConversionSeries.CONVERSIONS_COUNTER
        conversion_df = self._get_calculated_conversion_df(
            series_to_calculate=series_to_calculate,
            data=data,
            name=name,
            partition=partition,
        )

        series = conversion_df[series_to_calculate.private_name]
        return (
            series
            .copy_override(name=series_to_calculate.public_name)
            .copy_override_type(bach.SeriesInt64)
        )

    def conversion_count(self, *args, **kwargs):
        raise NotImplementedError('function is renamed please use `conversions_in_time`')

    @use_only_required_objectiv_series(
        required_series=['session_id', 'moment', 'event_type'],
        include_series_from_params=['partition'],
    )
    def conversions_in_time(self,
                            data: bach.DataFrame,
                            name: str,
                            partition: str = 'session_id') -> bach.SeriesInt64:
        """
        Counts the number of time a user is converted at a moment in time given a partition (ie 'session_id'
        or 'user_id').

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param name: the name of the conversion to label as set in
            :py:attr:`ModelHub.conversion_events`.
        :param partition: the partition over which the number of conversions are counted. Can be any column
            in ``data``.
        :returns: :py:class:`bach.SeriesInt64` with the same index as ``data``.
        """
        series_to_calculate = _CalculatedConversionSeries.CONVERSIONS_IN_TIME
        conversion_df = self._get_calculated_conversion_df(
            series_to_calculate=series_to_calculate,
            data=data,
            name=name,
            partition=partition,
        )

        series = conversion_df[series_to_calculate.private_name]
        return (
            series
            .copy_override(name=series_to_calculate.public_name)
            .copy_override_type(bach.SeriesInt64)
        )

    @use_only_required_objectiv_series(
        required_series=['session_id', 'session_hit_number', 'moment', 'event_type'],
        include_series_from_params=['partition'],
    )
    def pre_conversion_hit_number(
        self,
        data: bach.DataFrame,
        name: str,
        partition: str = 'session_id',
    ) -> bach.SeriesInt64:
        """
        Returns a count backwards from the first conversion, given the partition. I.e. first hit before
        converting is 1, second hit before converting 2, etc. Returns None if there are no conversions
        in the partition or after the first conversion.

        :param data: :py:class:`bach.DataFrame` to apply the method on.
        :param name: the name of the conversion to label as set in
            :py:attr:`ModelHub.conversion_events`.
        :param partition: the partition over which the number of conversions are counted. Can be any column
            in ``data``.
        :returns: :py:class:`bach.SeriesInt64` with the same index as ``data``.
        """

        series_to_calculate = _CalculatedConversionSeries.PRE_CONVERSION_HIT_NUMBER
        conversion_df = self._get_calculated_conversion_df(
            series_to_calculate=series_to_calculate,
            data=data,
            name=name,
            partition=partition,
        )

        series = conversion_df[series_to_calculate.private_name]

        return (
            series.copy_override(name=series_to_calculate.public_name).
            copy_override_type(bach.SeriesInt64)
        )

    @staticmethod
    def _check_conversion_dependencies(
        data: bach.DataFrame, series_to_calculate: _CalculatedConversionSeries
    ) -> None:
        """
        Helper function that checks if dataframe contains all dependent series required
        for calculating series_to_calculate
        """
        if series_to_calculate not in _DEPENDENCIES_PER_CONVERSION_SERIES:
            return None

        for dep in _DEPENDENCIES_PER_CONVERSION_SERIES[series_to_calculate]:
            if dep.private_name not in data.data_columns:
                raise Exception(f'{dep} is required for calculating {series_to_calculate} series')

    def _get_calculated_conversion_df(
        self,
        *,
        series_to_calculate: _CalculatedConversionSeries,
        data: bach.DataFrame,
        **kwargs,
    ) -> bach.DataFrame:
        """
        Generates a dataframe containing all required conversion series based on series_to_calculate param.
        All new series will contain private names.
        """
        dependencies = _DEPENDENCIES_PER_CONVERSION_SERIES.get(series_to_calculate, [])
        variables_to_calc = dependencies + [series_to_calculate]

        # independent variables
        if _CalculatedConversionSeries.IS_CONVERSION_EVENT in variables_to_calc:
            data = self._calculate_is_conversion_event(data, **kwargs)

        # dependent variables (order matters)
        if _CalculatedConversionSeries.CONVERSIONS_IN_TIME in variables_to_calc:
            data = self._calculate_conversions_in_time(data, **kwargs)

        if _CalculatedConversionSeries.CONVERSIONS_COUNTER in variables_to_calc:
            data = self._calculate_conversions_counter(data, **kwargs)

        if _CalculatedConversionSeries.PRE_CONVERSION_HIT_NUMBER in variables_to_calc:
            data = self._calculate_pre_conversion_hit_number(data, **kwargs)

        calculated_series = [var.private_name for var in variables_to_calc]
        return data[calculated_series]

    def _calculate_is_conversion_event(self, data: bach.DataFrame, name: str, **kwargs) -> bach.DataFrame:
        """ For documentation, see function :meth:`Map.is_conversion_event()` """

        series_to_calculate = _CalculatedConversionSeries.IS_CONVERSION_EVENT
        self._check_conversion_dependencies(data, series_to_calculate)

        if name not in self._mh._conversion_events:
            raise KeyError(f"Key {name} is not labeled as a conversion")

        conversion_stack, conversion_event = self._mh._conversion_events[name]

        if conversion_stack is None:
            series = data.event_type == conversion_event
        elif conversion_event is None:
            series = conversion_stack.json.get_array_length() > 0
        else:
            series = ((conversion_stack.json.get_array_length() > 0) & (data.event_type == conversion_event))

        data[series_to_calculate.private_name] = series
        return data

    def _calculate_conversions_in_time(
        self, data: bach.DataFrame, partition: str, *args, **kwargs,
    ) -> bach.DataFrame:
        """ For documentation, see function :meth:`Map.conversions_in_time()` """
        series_to_calculate = _CalculatedConversionSeries.CONVERSIONS_IN_TIME
        self._check_conversion_dependencies(data, series_to_calculate)

        is_conversion_event_column = _CalculatedConversionSeries.IS_CONVERSION_EVENT.private_name
        column_name_to_add_conversion_counter = '__conversion_counter'
        data[column_name_to_add_conversion_counter] = 0
        data.loc[data[is_conversion_event_column], column_name_to_add_conversion_counter] = 1

        window = data.sort_values([partition, 'moment']).groupby(partition).window()

        series = (
            data[column_name_to_add_conversion_counter]
            .copy_override_type(bach.SeriesInt64).sum(window)
        )
        data[series_to_calculate.private_name] = series

        # conversions_in_time requires materialization due to window
        return data.materialize(f'{series_to_calculate}_calculation')

    def _calculate_conversions_counter(
        self, data: bach.DataFrame, partition: str, **kwargs,
    ) -> bach.DataFrame:
        """ For documentation, see function :meth:`Map.conversions_counter()` """

        series_to_calculate = _CalculatedConversionSeries.CONVERSIONS_COUNTER
        self._check_conversion_dependencies(data, series_to_calculate)

        conversions_in_time_column = _CalculatedConversionSeries.CONVERSIONS_IN_TIME.private_name
        window = data.groupby(partition).window(end_boundary=WindowFrameBoundary.FOLLOWING)
        data[series_to_calculate.private_name] = window[conversions_in_time_column].max()

        # conversions_counter requires materialization due to window
        return data.materialize(node_name=f'{series_to_calculate}_calculation')

    def _calculate_pre_conversion_hit_number(
        self, data: bach.DataFrame, partition: str, **kwargs,
    ) -> bach.DataFrame:
        """ For documentation, see function :meth:`Map.pre_conversion_hit_number()` """

        series_to_calculate = _CalculatedConversionSeries.PRE_CONVERSION_HIT_NUMBER
        self._check_conversion_dependencies(data, series_to_calculate)

        conversions_in_time_column = _CalculatedConversionSeries.CONVERSIONS_IN_TIME.private_name

        # sort all windows by session id and hit number
        sort_by = ['session_id', 'session_hit_number']

        window = data.sort_values(sort_by, ascending=False).groupby(partition).window()
        max_number_of_conversions = data[conversions_in_time_column].max(window)
        data['__is_converted'] = True
        data.loc[max_number_of_conversions == 0, '__is_converted'] = False

        data.materialize(inplace=True)

        window = data.sort_values(sort_by, ascending=False).groupby(partition).window()

        # number all rows except where __is_converted is NULL and _conversions == 0
        pch_mask = (data['__is_converted']) & (data[conversions_in_time_column] == 0)

        calc_series_name = series_to_calculate.private_name
        data[calc_series_name] = 1
        data.loc[~pch_mask, calc_series_name] = None

        pre_conversion_hit_number = (
            data[calc_series_name].astype('int64')
            .copy_override_type(bach.SeriesInt64)  # None is parsed as string
        )
        data[calc_series_name] = pre_conversion_hit_number.sum(window)

        # pre_conversion_hit_number requires materialization due to window
        return data.materialize(node_name=f'{series_to_calculate}_calculation')



"""
Copyright 2022 Objectiv B.V.
"""
from abc import abstractmethod
from typing import Dict, Any

import bach
from bach import get_series_type_from_dtype


class BaseDataPipeline:
    """
    Abstract class that specifies the engine and table from where the pipeline will extract the data.
    Child classes are in charge of implementing:
     - _get_pipeline_result method, which is in charge of generating the correct bach DataFrame
     - validate_pipeline_result method, which is in charge of verifying that the
       pipeline is generating the expected result.
     - result_series_dtypes property, which contains the mapping between the final series
        generated by the pipeline and expected dtype.
    """

    def __call__(self, **kwargs) -> bach.DataFrame:
        result = self._get_pipeline_result(**kwargs)
        self.validate_pipeline_result(result)
        return result

    @classmethod
    @abstractmethod
    def validate_pipeline_result(cls, result: bach.DataFrame) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _get_pipeline_result(self, **kwargs) -> bach.DataFrame:
        raise NotImplementedError()

    @property
    def result_series_dtypes(self) -> Dict[str, str]:
        """
        Mapping between series generated by pipeline with its respective dtype.
        """
        raise NotImplementedError()

    def _convert_dtypes(self, df: bach.DataFrame) -> bach.DataFrame:
        """
        Helper function that converts each result series to its correct dtype,
        this way we ensure the pipeline is returning the dtypes Modelhub is expecting.

        Returns a bach DataFrame
        """
        df_cp = df.copy()
        for col, dtype in self.result_series_dtypes.items():
            if col not in df_cp.data:
                continue

            df_cp[col] = df_cp[col].astype(dtype)

        return df_cp

    def _validate_data_dtypes(
        self, expected_dtypes: Dict[str, Any], current_dtypes: Dict[str, Any],
    ) -> None:
        """
        Compares dtypes mapping, based on:
            - all keys in expected_dtypes must exist in current_dtypes
            - Series type for expected dtype must be a parent class or same class from
              its respective current dtype. For example:
                expected_dtype = 'json',             current_dtype = 'objectiv_global_context'
                expected_series_type = SeriesJson,   current_series_type = SeriesGlobalContexts

              Therefore, current_dtype is valid.
        """
        missing_columns = set(expected_dtypes.keys()) - set(current_dtypes.keys())
        if missing_columns:
            raise KeyError(
                f'{self.__class__.__name__} expects mandatory columns: {expected_dtypes.keys()}, '
                f'missing: {",".join(missing_columns)}.'
            )

        for expected_key, expected_dtype in expected_dtypes.items():
            current_dtype = current_dtypes[expected_key]

            expected_series_type = get_series_type_from_dtype(expected_dtype)
            current_series_type = get_series_type_from_dtype(current_dtype)
            if (
                expected_series_type != current_series_type
                and not issubclass(current_series_type, expected_series_type)
            ):
                raise ValueError(
                    f'"{expected_key}" must be {expected_dtype} dtype, got {current_dtype}'
                )



"""
Copyright 2021 Objectiv B.V.
"""
from enum import Enum
from typing import Dict, Optional

import bach

from modelhub.util import (
    ObjectivSupportedColumns, get_supported_dtypes_per_objectiv_column, check_objectiv_dataframe
)
from modelhub.pipelines.base_pipeline import BaseDataPipeline


class _BaseCalculatedSessionSeries(Enum):
    IS_START_OF_SESSION = 'is_start_of_session'
    SESSION_START_ID = 'session_start_id'
    SESSION_COUNT = 'is_one_session'


class SessionizedDataPipeline(BaseDataPipeline):
    """
    Pipeline in charge of calculating Objectiv sessionized columns.
    This pipeline is dependent on the result from ExtractedContextsPipeline, therefore it expects that
    the result from the latter is generated correctly.

    The steps followed in this pipeline are the following:
        1. _validate_extracted_context_df: Validates if provided DataFrame contains event_id, user_id and
            moment series.
        2. _calculate_base_session_series: Calculates `is_start_of_session`, `session_start_id` and
            `is_one_session` series.
        3. _calculate_objectiv_session_series: Calculates final sessionized series:
            `session_id` and `session_hit_number`
        4. _convert_dtypes: Will convert all required sessionized series to their correct dtype

    Final bach DataFrame will be later validated, it must include:
        - session_id and session_hit_number series
        - correct dtypes for sessionized series
    """

    def __init__(self, session_gap_seconds: int):
        self._session_gap_seconds = session_gap_seconds

    def _get_pipeline_result(
        self, extracted_contexts_df: Optional[bach.DataFrame] = None, **kwargs
    ) -> bach.DataFrame:
        """
        Contains steps for calculating sessionized series for provided dataframe.
        :param extracted_contexts_df: bach DataFrame containing `user_id`, `event_id` and `moment` series.

        returns a bach DataFrame with session_id and session_hit_number series
            and all series from provided extracted_contexts_df.
        """
        if not extracted_contexts_df:
            raise ValueError(f'{self.__class__.__name__} requires dataframe with extracted contexts.')

        context_df = extracted_contexts_df.copy()
        self._validate_extracted_context_df(context_df)

        # calculate series that are needed for the final result
        sessionized_df = self._calculate_base_session_series(context_df)

        # adds required objectiv session series
        sessionized_df = self._calculate_objectiv_session_series(sessionized_df)

        sessionized_df = self._convert_dtypes(df=sessionized_df)

        final_columns = context_df.data_columns + ObjectivSupportedColumns.get_sessionized_columns()
        return sessionized_df[final_columns]

    def _validate_extracted_context_df(self, df: bach.DataFrame) -> None:
        # make sure the result has AT LEAST the following series
        # even though the ExtractedContextsPipeline already validates series
        with_identity_resolution = (
            ObjectivSupportedColumns.IDENTITY_USER_ID.value in df.data_columns
        )
        supported_dtypes = get_supported_dtypes_per_objectiv_column(
            with_identity_resolution=with_identity_resolution
        )
        expected_context_columns = [
            ObjectivSupportedColumns.EVENT_ID.value,
            ObjectivSupportedColumns.USER_ID.value,
            ObjectivSupportedColumns.MOMENT.value,
        ]

        self._validate_data_dtypes(
            expected_dtypes={col: supported_dtypes[col] for col in expected_context_columns},
            current_dtypes=df.dtypes,
        )

    @classmethod
    def validate_pipeline_result(cls, result: bach.DataFrame) -> None:
        """
        Checks if we are returning all sessionized series with respective dtype.
        """
        check_objectiv_dataframe(
            result,
            columns_to_check=ObjectivSupportedColumns.get_sessionized_columns(),
            check_dtypes=True,
            infer_identity_resolution=True,
        )

    @property
    def result_series_dtypes(self) -> Dict[str, str]:
        sessionized_columns = ObjectivSupportedColumns.get_sessionized_columns()
        return {
            col: dtype for col, dtype in get_supported_dtypes_per_objectiv_column().items()
            if col in sessionized_columns
        }

    def _calculate_base_session_series(self, df: bach.DataFrame) -> bach.DataFrame:
        """
        Calculates each series required for calculating the final sessionized series.

        Series to calculate:
            - is_start_of_session: boolean series that defines if an event/row
                is the start of a user's session
            - session_start_id: Session number in the whole dataset,
                only rows where is_start_of_session is True are numbered.
            - is_one_session: amount of observed session starts before current session start

        Returns a bach DataFrame
        """
        sessionized_df = df.copy()

        is_session_start_series = self._calculate_session_start(sessionized_df, self._session_gap_seconds)
        sessionized_df[is_session_start_series.name] = is_session_start_series
        # materialize since rest of series are dependant and it uses a window function
        sessionized_df = sessionized_df.materialize(node_name='session_starts')

        session_start_id_series = self._calculate_session_start_id(sessionized_df)
        sessionized_df[session_start_id_series.name] = session_start_id_series

        session_count_series = self._calculate_session_count(sessionized_df)
        sessionized_df[session_count_series.name] = session_count_series
        return sessionized_df.materialize(node_name='session_id_and_count')

    @staticmethod
    def _calculate_objectiv_session_series(df: bach.DataFrame) -> bach.DataFrame:
        """
        Calculates all Sessionized Objectiv series expected by Modelhub. df MUST contain all
        calculated series from _calculate_base_session_series step.

        Series to calculate:
           - session_id: correct session_id for all rows with the same value for is_one_session
           - session_hit_number: event's number in respective session
        Returns a bach DataFrame
        """
        sort_by = [ObjectivSupportedColumns.MOMENT.value, ObjectivSupportedColumns.EVENT_ID.value]
        group_by = [_BaseCalculatedSessionSeries.SESSION_COUNT.value]
        window = df.sort_values(by=sort_by).groupby(group_by).window()

        df_cp = df.copy()
        session_start_id_series = df_cp[_BaseCalculatedSessionSeries.SESSION_START_ID.value]

        session_id_series_name = ObjectivSupportedColumns.SESSION_ID.value
        df_cp[session_id_series_name] = session_start_id_series.copy_override(
            expression=session_start_id_series.window_first_value(window=window).expression,
            name=session_id_series_name,
        )

        session_hit_number_series_name = ObjectivSupportedColumns.SESSION_HIT_NUMBER.value
        df_cp[session_hit_number_series_name] = session_start_id_series.copy_override(
            expression=session_start_id_series.window_row_number(window=window).expression,
            name=session_hit_number_series_name,
        )

        return df_cp.materialize(node_name='objectiv_sessionized_data')

    @staticmethod
    def _calculate_session_start(df: bach.DataFrame, session_gap_seconds: int) -> bach.SeriesBoolean:
        """
        Generates is_session_start series which determines if an event is the first event of a user's session.

        This is performed by:
          1. Grouping events by user ids and ordering by moment and event id

          2. Calculate the amount of time between previous and current event.
            (event_lapsed_time = events[n][moment] - events[n-1][moment])

          3. How to determine if the event is start of a session:
            a) If the row is the first observed event for the user's entire history, then by default it is a
                session start.

            b) Identify which event is the end of the session by comparing time differences.
                If the event_lapsed time is greater than session_gap_seconds, then we can deduct the current
                event is a session start.

        Returns bach SeriesBoolean
        """
        sort_by = [ObjectivSupportedColumns.MOMENT.value, ObjectivSupportedColumns.EVENT_ID.value]
        group_by = [ObjectivSupportedColumns.USER_ID.value]
        window = df.sort_values(by=sort_by).groupby(by=group_by).window()

        moment_series = df[ObjectivSupportedColumns.MOMENT.value]

        previous_moment_series = moment_series.window_lag(window=window)
        previous_moment_series = previous_moment_series.copy_override_type(bach.SeriesTimestamp)

        event_lapsed_time = moment_series - previous_moment_series
        assert isinstance(event_lapsed_time, bach.SeriesTimedelta)  # help mypy

        result_series_name = _BaseCalculatedSessionSeries.IS_START_OF_SESSION.value
        df_cp = df.copy()
        # let's assume all events are session starts
        df_cp[result_series_name] = True

        session_gap_mask = event_lapsed_time.dt.total_seconds <= session_gap_seconds
        # set None to those events with lapse time less than or equal to the session_gap_seconds
        # we can assume those are part of a same session
        # We use None instead of False because _calculate_session_count will count the values
        # from the resultant series, therefore we want to ignore non session start events
        df_cp.loc[session_gap_mask, result_series_name] = bach.SeriesBoolean.from_value(
            base=df_cp,
            value=None,
            name=result_series_name,
        )
        return df_cp[result_series_name].copy_override_type(bach.SeriesBoolean)

    @staticmethod
    def _calculate_session_start_id(df: bach.DataFrame) -> bach.SeriesInt64:
        """
        Calculates session_start_id by numbering each event that is a session start.
        Returns a bach SeriesInt64
        """
        sort_by = [ObjectivSupportedColumns.MOMENT.value, ObjectivSupportedColumns.EVENT_ID.value]
        group_by = [_BaseCalculatedSessionSeries.IS_START_OF_SESSION.value]
        window = df.sort_values(by=sort_by).groupby(by=group_by).window()

        is_start_session_series = df[_BaseCalculatedSessionSeries.IS_START_OF_SESSION.value]
        # group all rows by is_start_of_session and number each event
        session_start_id = is_start_session_series.window_row_number(window=window)
        session_start_id = session_start_id.copy_override_type(bach.SeriesInt64)

        result_series_name = _BaseCalculatedSessionSeries.SESSION_START_ID.value
        df_cp = df.copy()
        df_cp[result_series_name] = bach.SeriesInt64.from_value(
            base=df_cp, value=None, name=result_series_name,
        )
        # consider only events that are start of session
        df_cp.loc[is_start_session_series, result_series_name] = session_start_id
        return df_cp[result_series_name]

    @staticmethod
    def _calculate_session_count(df: bach.DataFrame) -> bach.SeriesInt64:
        """
        generates a unique number for each session, but not in the right order,
        by calculating the amount of observed session starts before current session start.
        For example:
           event_id  is_start_of_session   is_one_session
              1            True                1
              2            None                1
              3            True                2
              4            True                3
              5            None                3

        Returns a bach SeriesInt64
        """
        sort_by = [
            ObjectivSupportedColumns.USER_ID.value,
            ObjectivSupportedColumns.MOMENT.value,
            ObjectivSupportedColumns.EVENT_ID.value,
        ]
        window = df.sort_values(by=sort_by).groupby().window()

        start_session_series = df[_BaseCalculatedSessionSeries.IS_START_OF_SESSION.value]
        session_count_series = start_session_series.count(partition=window)
        return session_count_series.copy_override(
            name=_BaseCalculatedSessionSeries.SESSION_COUNT.value,
        ).copy_override_type(bach.SeriesInt64)



"""
Copyright 2021 Objectiv B.V.
"""
import json
import os

from typing import List, Generic, TypeVar
# added for metabase export
import requests

from bach import DataFrame, Series, SeriesString, SeriesJson
from bach.expression import Expression, quote_string, quote_identifier, join_expressions
from bach.series.series_json import JsonAccessor
from bach.types import register_dtype
from sql_models.util import is_postgres, DatabaseNotSupportedException, is_bigquery


TSeriesJson = TypeVar('TSeriesJson', bound='SeriesJson')


class ObjectivStack(JsonAccessor, Generic[TSeriesJson]):
    """
    Specialized JsonAccessor that has functions to work on SeriesJsons whose data consists of an array
    of objects.
    """

    def get_from_context_with_type_series(self, type: str, key: str, dtype='string'):
        """
        .. _get_from_context_with_type_series:

        Returns the value of `key` from the first context in an Objectiv stack where `_type` matches `type`.

        :param type: the _type to search for in the contexts of the stack.
        :param key: the value of the key to return of the context with matching type.
        :param dtype: the dtype of the series to return.
        :returns: a series of type `dtype`
        """
        dialect = self._series_object.engine.dialect
        if is_postgres(dialect):
            return self._postgres_get_from_context_with_type_series(type, key, dtype)
        if is_bigquery(dialect):
            return self._bigquery_get_from_context_with_type_series(type, key, dtype)
        raise DatabaseNotSupportedException(dialect)

    def _postgres_get_from_context_with_type_series(self, type: str, key: str, dtype='string'):
        dialect = self._series_object.engine.dialect
        expression_str = f'''
        jsonb_path_query_first({{}},
        \'$[*] ? (@._type == $type)\',
        \'{{"type":{quote_identifier(dialect, type)}}}\') ->> {{}}'''
        expression = Expression.construct(
            expression_str,
            self._series_object,
            Expression.string_value(key)
        )
        return self._series_object.copy_override_dtype(dtype).copy_override(expression=expression)

    def _bigquery_get_from_context_with_type_series(self, type: str, key: str, dtype='string'):
        select_ctx_expression = Expression.construct(
            '''(
              select first_value(ctx) over (order by pos)
              from unnest(json_query_array({}, '$')) as ctx with offset as pos
              where json_value(ctx, '$."_type"') = {}
            )''',
            self._series_object,
            Expression.string_value(type)
        )
        ctx_series = self._series_object.copy_override(expression=select_ctx_expression)
        as_str = dtype == 'string'
        value_series = ctx_series.json.get_value(key=key, as_str=as_str)
        return value_series.copy_override_dtype(dtype)


@register_dtype(value_types=[], override_registered_types=True)
class SeriesGlobalContexts(SeriesJson):
    """
    Objectiv Global Contexts series. This series type contains functionality specific to the Objectiv Global
    Contexts.
    """
    dtype = 'objectiv_global_context'

    class GlobalContexts(ObjectivStack):
        @property
        def cookie_id(self):
            """
            .. _gc_cookie_id:

            Returns cookie id from the global contexts.
            """
            return self.get_from_context_with_type_series("CookieIdContext", "cookie_id")

        @property
        def user_agent(self):
            """
            .. _gc_user_agent:

            Returns user agent string from the global contexts.
            """
            return self.get_from_context_with_type_series("HttpContext", "user_agent")

        @property
        def application(self):
            """
            .. _gc_application:

            Returns application id from the global contexts.
            """
            return self.get_from_context_with_type_series("ApplicationContext", "id")

    @property
    def objectiv(self):
        """
        Accessor for Objectiv stack data. All methods of :py:attr:`json` can also be accessed with this
        accessor. Same as :py:attr:`obj`

        .. autoclass:: modelhub.series.ObjectivStack
            :members:
            :noindex:

        """
        return ObjectivStack(self)

    @property
    def obj(self):
        """
        Accessor for Objectiv stack data. All methods of :py:attr:`json` can also be accessed with this
        accessor. Same as :py:attr:`objectiv`

        .. autoclass:: modelhub.series.ObjectivStack
            :members:
            :noindex:

        """
        return ObjectivStack(self)

    @property
    def global_contexts(self):
        """
        Accessor for Objectiv global context data. All methods of :py:attr:`json` and :py:attr:`objectiv` can
        also be accessed with this accessor. Same as :py:attr:`gc`

        .. autoclass:: modelhub.series.SeriesGlobalContexts.GlobalContexts
            :members:

        """
        return self.GlobalContexts(self)

    @property
    def gc(self):
        """
        Accessor for Objectiv global context data. All methods of :py:attr:`json` and :py:attr:`objectiv` can
        also be accessed with this accessor. Same as :py:attr:`global_contexts`

        .. autoclass:: modelhub.series.SeriesGlobalContexts.GlobalContexts
            :members:
            :noindex:

        """
        return self.GlobalContexts(self)


@register_dtype([], override_registered_types=True)
class SeriesLocationStack(SeriesJson):
    """
    Objectiv Location Stack series. This series type contains functionality specific to the Objectiv Location
    Stack.
    """
    dtype = 'objectiv_location_stack'

    class LocationStack(ObjectivStack):
        @property
        def navigation_features(self):
            """
            .. _ls_navigation_features:

            Returns the navigation stack from the location stack.
            """
            # type ignore, as mypy doesn't like a dict in a Slice
            return self[{'_type': 'NavigationContext'}: None]  # type: ignore

        @property
        def feature_stack(self) -> 'SeriesLocationStack':
            """
            .. _ls_feature_stack:

            Returns the feature stack from the location stack. The context objects only contain the `_type`
            and a `id` key.
            """
            keys = ['_type', 'id']
            engine = self._series_object.engine
            if is_postgres(engine):
                return self._postgres_filter_keys_of_dicts(keys)
            if is_bigquery(engine):
                return self._bigquery_filter_keys_of_dicts(keys)
            raise DatabaseNotSupportedException(engine)

        def _postgres_filter_keys_of_dicts(self, keys: List[str]) -> 'TSeriesJson':
            """
            Return a new Series, that consists of the same top level array, but with all the sub-dictionaries
            having only their original fields if that field is listed in the keys parameter.

            Postgres only.
            """
            jsonb_build_object_str = [f"{quote_string(self._series_object.engine, key)}" for key in keys]
            expression_str = f'''(
                select jsonb_agg((
                    select json_object_agg(items.key, items.value)
                    from jsonb_each(objects.value) as items
                    where items.key in ({", ".join(jsonb_build_object_str)})))
                from jsonb_array_elements({{}}) as objects)
            '''
            expression = Expression.construct(
                expression_str,
                self._series_object
            )
            return self._series_object.copy_override(expression=expression)

        def _bigquery_filter_keys_of_dicts(self, keys: List[str]) -> 'TSeriesJson':
            """
            Return a new Series, that consists of the same top level array, but with all the sub-dictionaries
            having only the fields that are listed in the keys parameter. If the field didn't previously
            exist then it will have the value NULL.

            The fields of the sub-dictionaries MUST be scalar values (i.e. string, number, or boolean).
            Bigquery only.
            """
            # We unnest the json-array, and then for every json object in the array we build a copy of that
            # json object, but with only the keys that are listed in the `keys` variable.
            object_entries_expressions = []
            for i, key in enumerate(keys):
                if '"' in key:
                    raise ValueError(f'key values containing double quotes are not supported. key: {key}')
                # Using JSON_VALUE here means this only works with scalar, but we warn the user in the
                # docstring of this.
                entry_expr = Expression.construct(
                    '''JSON_VALUE(item, '$."{}"') as {}''',
                    Expression.raw(key),
                    Expression.identifier(key)
                )
                object_entries_expressions.append(entry_expr)
            expression = Expression.construct(
                '''TO_JSON_STRING(ARRAY(
                    SELECT STRUCT({})
                    FROM UNNEST (JSON_QUERY_ARRAY({}, '$')) AS item WITH OFFSET AS pos
                    ORDER BY pos))''',
                join_expressions(object_entries_expressions),
                self._series_object
            )
            return self._series_object.copy_override(expression=expression)

        @property
        def nice_name(self) -> 'SeriesString':
            """
            .. _ls_nice_name:

            Returns a nice name for the location stack. This is a human readable name for the data in the
            feature stack.
            """
            engine = self._series_object.engine
            if is_postgres(engine):
                expression = self._postgres_nice_name()
            elif is_bigquery(engine):
                expression = self._bigquery_nice_name()
            else:
                raise DatabaseNotSupportedException(engine)
            return self._series_object.copy_override_type(SeriesString).copy_override(expression=expression)

        def _bigquery_nice_name(self) -> Expression:
            # last_element_nice_expr turns something with _type='SectionContext' and id='section_id'
            # into something like 'Section: section_id'
            last_element = self._series_object.json[-1]
            last_element_nice_expr = Expression.construct(
                """
                    REGEXP_REPLACE(
                        REGEXP_REPLACE({}, '([a-z])([A-Z])', r'\\1 \\2'),
                        ' Context$',
                        ''
                    ) || ': ' || {}
                """,
                last_element.json.get_value('_type', as_str=True),
                last_element.json.get_value('id', as_str=True)
            )

            # For all other elements in the stack we do something similar as we did with the last element
            # in last_element_nice_expr. But we prepend 'Located at', and couple the elements with ' => '.
            # Example output value for other_elements_expr:
            #   'located at Web Document: #document => Section: x => Section: y'

            # element_json allows us to reuse code from the accessor without duplicating. 'element' will be
            # set by iterating over the json array in the query below.
            element_json = self._series_object.copy_override(expression=Expression.construct('element'))
            other_elements_expr = Expression.construct(
                """
                    CASE WHEN {} > 1
                    THEN ' located at ' || (SELECT STRING_AGG(
                                    REGEXP_REPLACE(
                                        REGEXP_REPLACE({}, '([a-z])([A-Z])', r'\\1 \\2'),
                                        ' Context$',
                                        ''
                                    ) || ': ' || {},
                                    ' => '
                                )
                                FROM UNNEST (JSON_QUERY_ARRAY({}, '$')) AS element WITH OFFSET AS pos
                                WHERE pos < ({} - 1)
                            )
                    ELSE ''
                    END
                """,
                self._series_object.json.get_array_length(),
                element_json.json.get_value('_type', as_str=True),
                element_json.json.get_value('id', as_str=True),
                self._series_object,
                self._series_object.json.get_array_length(),
            )
            return Expression.construct('({}) || ({})', last_element_nice_expr, other_elements_expr)

        def _postgres_nice_name(self) -> Expression:
            expression = Expression.construct(
                f"""(
                    select string_agg(
                            replace(
                                regexp_replace(value ->> '_type', '([a-z])([A-Z])', '\\1 \\2', 'g'),
                                ' Context',
                                ''
                            ) || ': ' || (value ->> 'id'),
                            ' => ')
                    from jsonb_array_elements({{}}) with ordinality
                    where ordinality = jsonb_array_length({{}})
                ) || (
                    case when jsonb_array_length({{}}) > 1
                         then ' located at ' || (select string_agg(
                                replace(
                                    regexp_replace(value ->> '_type', '([a-z])([A-Z])', '\\1 \\2', 'g'),
                                    ' Context',
                                    ''
                                ) || ': ' || (value ->> 'id'),
                                ' => ')
                            from jsonb_array_elements({{}}) with ordinality
                            where ordinality < jsonb_array_length({{}})
                        )
                        else '' end
                )""",
                self._series_object,
                self._series_object,
                self._series_object,
                self._series_object,
                self._series_object
            )
            return expression

    @property
    def objectiv(self):
        """
        Accessor for Objectiv stack data. All methods of :py:attr:`json` can also be accessed with this
        accessor. Same as :py:attr:`obj`

        .. autoclass:: modelhub.series.ObjectivStack
            :members:
            :noindex:

        """
        return ObjectivStack(self)

    @property
    def obj(self):
        """
        Accessor for Objectiv stack data. All methods of :py:attr:`json` can also be accessed with this
        accessor. Same as :py:attr:`objectiv`

        .. autoclass:: modelhub.series.ObjectivStack
            :members:
            :noindex:

        """
        return ObjectivStack(self)

    @property
    def location_stack(self):
        """
        Accessor for Objectiv location stack data. All methods of :py:attr:`json` and :py:attr:`objectiv`
        can also be accessed with this accessor. Same as :py:attr:`ls`

        .. autoclass:: modelhub.series.SeriesLocationStack.LocationStack
            :members:

        """
        return self.LocationStack(self)

    @property
    def ls(self):
        """
        Accessor for Objectiv location stack data. All methods of :py:attr:`json` and :py:attr:`objectiv` can
        also be accessed with this accessor. Same as :py:attr:`location_stack`

        .. autoclass:: modelhub.series.SeriesLocationStack.LocationStack
            :members:
            :noindex:

        """
        return self.LocationStack(self)


class MetaBaseException(Exception):
    pass


class MetaBase:

    _session_id = None

    # config per model
    config = {
        'default': {
            'display': 'bar',
            'name': 'Generic / default graph',
            'description': 'This is a generic graph',
            'result_metadata': [],
            'dimensions': [],
            'metrics': []
        },
        'unique_users': {
            'display': 'line',
            'name': 'Unique Users',
            'description': 'Unique Users',
            'result_metadata': [],
            'dimensions': ['date'],
            'metrics': ['count']
        },
        'unique_sessions': {
            'display': 'bar',
            'name': 'Unique Sessions',
            'description': 'Unique sessions from Model Hub',
            'result_metadata': [],
            'dimensions': ['date'],
            'metrics': ['count']
        }
    }

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 url: str = None,
                 database_id: int = None,
                 dashboard_id: int = None,
                 collection_id: int = None,
                 web_url: str = None):
        if username:
            self._username = username
        else:
            self._username = os.getenv('METABASE_USERNAME', 'objectiv')

        if password:
            self._password = password
        else:
            self._password = os.getenv('METABASE_PASSWORD', '')

        if database_id:
            self._database_id = database_id
        else:
            self._database_id = int(os.getenv('METABASE_DATABASE_ID', 1))

        if dashboard_id:
            self._dashboard_id = dashboard_id
        else:
            self._dashboard_id = int(os.getenv('METABASE_DASHBOARD_ID', 1))

        if collection_id:
            self._collection_id = collection_id
        else:
            self._collection_id = int(os.getenv('METABASE_COLLECTION_ID', 0))

        if url:
            self._url = url
        else:
            self._url = os.getenv('METABASE_URL', '2')

        if web_url:
            self._web_url = web_url
        else:
            self._web_url = os.getenv('METABASE_WEB_URL', self._url)

        # config by calling dataframe / model
        self._df = None
        self._config = None

    def _get_new_session_id(self) -> str:
        data = json.dumps({'username': self._username, 'password': self._password})
        headers = {'Content-Type': 'application/json'}
        response = requests.post(f'{self._url}/api/session', data=data, headers=headers)

        if response.status_code != 200:
            raise MetaBaseException(f'Session ID request failed with code: {response.status_code}')

        response_json = response.json()

        if 'id' in response_json:
            return response_json['id']
        else:
            raise KeyError('Could not find id in JSON response from MetaBase')

    def _get_session_id(self):
        if MetaBase._session_id is None:
            MetaBase._session_id = self._get_new_session_id()

        return MetaBase._session_id

    def _do_request(self, url: str, data: dict = None, method='post') -> requests.Response:
        if data is None:
            data = {}
        headers = {
            'Content-Type': 'application/json',
            'X-Metabase-Session': self._get_session_id()
        }
        if method == 'get':
            response = requests.get(url, data=json.dumps(data), headers=headers)
        elif method == 'post':
            response = requests.post(url, data=json.dumps(data), headers=headers)
        elif method == 'put':
            response = requests.put(url, data=json.dumps(data), headers=headers)
        else:
            raise MetaBaseException(f'Unsupported method called: {method}')

        return response

    def add_update_card(self, df: DataFrame, config: dict) -> dict:
        data = {
            'collection_id': self._collection_id,
            'dataset_query': {
                'database': self._database_id,
                'native': {
                    'query': df.view_sql()
                },
                'type': 'native'
            },
            'description': config['description'],
            'display': config['display'],
            'name': config['name'],
            'result_metadata': config['result_metadata'],
            'visualization_settings': {
                'graph.dimensions': config['dimensions'],
                'graph.metrics': config['metrics']
            }
        }
        response = self._do_request(url=f'{self._url}/api/card', method='get')

        if response.status_code != 200:
            raise MetaBaseException(f'Failed to obtain list of existing cards with code: '
                                    f'{response.status_code}')

        # the default is to create a new card
        method = 'post'
        url = f'{self._url}/api/card'

        # but if we can find an existing card that matches
        # we update, rather than create
        for card in response.json():
            if card['description'] == config['description'] and \
                    card['name'] == config['name']:

                card_id = card['id']
                url = f'{self._url}/api/card/{card_id}'
                method = 'put'

        response = self._do_request(url=url, data=data, method=method)
        if response.status_code != 202:
            raise MetaBaseException(f'Failed to add card @ {url} with {data} (code={response.status_code})')

        response_json = response.json()
        if 'id' in response_json:
            card_id = response_json['id']
        else:
            raise MetaBaseException(f'No card ID in response {response_json}')

        dashboard_info = self.update_dashboard(card_id=card_id, dashboard_id=self._dashboard_id)

        return {
            'card': f'{self._web_url}/card/{card_id}',
            'dashboard': f'{self._web_url}/dashboard/{self._dashboard_id}-'
                         f'{dashboard_info["name"].lower().replace(" ", "-")}',
            'username': self._username,
            'password': self._password
        }

    def update_dashboard(self, card_id: int, dashboard_id: int):
        response = self._do_request(f'{self._url}/api/dashboard/{dashboard_id}', method='get')

        if response.status_code != 200:
            raise MetaBaseException(f'Failed to get cards list for dashboard {dashboard_id} '
                                    f'(code={response.status_code}')

        dashboard_info = response.json()
        # list of card_id's currently on the dashboard
        cards = [card['card']['id'] for card in dashboard_info['ordered_cards']]
        if card_id not in cards:

            url = f'{self._url}/api/dashboard/{dashboard_id}/cards'
            data = {'cardId': card_id}

            response = self._do_request(url=url, method='post', data=data)

            if response.status_code != 200:
                raise ValueError(f'Adding card to dashboard failed with code: {response.status_code}')
        return dashboard_info

    def to_metabase(self, df: DataFrame, model_type: str = None, config: dict = None):
        if isinstance(df, Series):
            df = df.to_frame()
        if not config:
            config = {}

        if model_type in MetaBase.config:
            card_config = MetaBase.config[model_type]
        else:
            card_config = MetaBase.config['default']

        card_config['dimensions'] = [k for k in df.index.keys()]
        card_config['metrics'] = [k for k in df.data.keys()]

        card_config.update(config)
        return self.add_update_card(df, card_config)



"""
Copyright 2021 Objectiv B.V.
"""
import operator
from functools import reduce

import bach
from typing import Optional, Any, Mapping, Dict, NamedTuple

from bach.types import StructuredDtype
from sql_models.constants import DBDialect
from sql_models.util import is_bigquery, is_postgres
from sqlalchemy.engine import Engine

from modelhub.util import (
    ObjectivSupportedColumns, get_supported_dtypes_per_objectiv_column, check_objectiv_dataframe
)
from modelhub.pipelines.base_pipeline import BaseDataPipeline


class TaxonomyColumnDefinition(NamedTuple):
    name: str
    dtype: Any


_DEFAULT_TAXONOMY_DTYPE_PER_FIELD = {
    '_type': bach.SeriesString.dtype,
    '_types': bach.SeriesJson.dtype,
    'global_contexts': bach.SeriesJson.dtype,
    'location_stack': bach.SeriesJson.dtype,
    'time': bach.SeriesInt64.dtype,
}

_PG_TAXONOMY_COLUMN = TaxonomyColumnDefinition(name='value', dtype=bach.SeriesJson.dtype)

_BQ_TAXONOMY_COLUMN = TaxonomyColumnDefinition(
    name='contexts_io_objectiv_taxonomy_1_0_0',  # change only when version is updated
    dtype=[
        {
            **_DEFAULT_TAXONOMY_DTYPE_PER_FIELD,
            'cookie_id': bach.SeriesUuid.dtype,
            'event_id': bach.SeriesUuid.dtype,
        },
    ],
)


def _get_taxonomy_column_definition(engine: Engine) -> TaxonomyColumnDefinition:
    """
    Returns the definition of the Objectiv taxonomy column per supported engine
    """
    if is_postgres(engine):
        return _PG_TAXONOMY_COLUMN

    if is_bigquery(engine):
        return _BQ_TAXONOMY_COLUMN

    raise Exception('define taxonomy definition for engine')


class ExtractedContextsPipeline(BaseDataPipeline):
    """
    Pipeline in charge of extracting Objectiv context columns from event data source.
    Based on the provided engine, it will perform the proper transformations for generating a bach DataFrame
    that contains all expected series to be used by Modelhub.

    The steps followed in this pipeline are the following:
        1. _get_initial_data: Gets the initial bach DataFrame containing all required context columns
            based on the engine.
        2. _process_taxonomy_data: Will extract required fields from taxonomy json, it will use
            _DEFAULT_TAXONOMY_DTYPE_PER_FIELD by default. If engine requires other fields,
            it should be specified to the pipeline.
        3. _apply_extra_processing: Applies extra processing if required, this is mainly based on the engine.
            For example, for BigQuery, the pipeline will generate moment and day series from `time` series
            extracted from the taxonomy.
        4. _convert_dtypes: Will convert all required context series to their correct dtype
        5. _apply_date_filter: Applies start_date and end_date filter, if required.

    Final bach DataFrame will be later validated, it must include:
        - all context series defined in ObjectivSupportedColumns
        - correct dtypes for context series
    """
    DATE_FILTER_COLUMN = ObjectivSupportedColumns.DAY.value

    # extra columns that are needed to be fetched, taxonomy column provided for the engine
    # is ALWAYS selected
    required_context_columns_per_dialect: Dict[DBDialect, Dict[str, StructuredDtype]] = {
        DBDialect.POSTGRES: {
            'event_id': bach.SeriesUuid.dtype,
            'day': bach.SeriesDate.dtype,
            'moment': bach.SeriesTimestamp.dtype,
            'cookie_id': bach.SeriesUuid.dtype,
        },
        DBDialect.BIGQUERY: {
            'collector_tstamp': bach.SeriesTimestamp.dtype,
        },
    }

    def __init__(self, engine: Engine, table_name: str):
        super().__init__()
        self._engine = engine
        self._table_name = table_name
        self._taxonomy_column = _get_taxonomy_column_definition(engine)

        # check if table has all required columns for pipeline
        dtypes = bach.from_database.get_dtypes_from_table(
            engine=self._engine,
            table_name=self._table_name,
        )
        self._validate_data_dtypes(
            expected_dtypes=dict(self._get_base_dtypes()),
            current_dtypes=dtypes,
        )

    def _get_pipeline_result(self, **kwargs) -> bach.DataFrame:
        """
        Calls all operations performed to the initial data and returns a bach DataFrame with the expected
        context series
        """
        context_df = self._get_initial_data()
        context_df = self._process_taxonomy_data(context_df)
        context_df = self._apply_extra_processing(context_df)

        context_df = self._convert_dtypes(df=context_df)
        context_df = self._apply_date_filter(df=context_df, **kwargs)

        context_columns = ObjectivSupportedColumns.get_extracted_context_columns()
        context_df = context_df[context_columns]
        return context_df.materialize(node_name='context_data')

    @property
    def result_series_dtypes(self) -> Dict[str, str]:
        context_columns = ObjectivSupportedColumns.get_extracted_context_columns()
        supported_dtypes = get_supported_dtypes_per_objectiv_column(with_identity_resolution=False)
        return {
            col: dtype for col, dtype in supported_dtypes.items()
            if col in context_columns
        }

    def _get_base_dtypes(self) -> Mapping[str, StructuredDtype]:
        """
        Returns mapping of series names and dtypes expected from initial data
        """
        db_dialect = DBDialect.from_engine(self._engine)
        return {
            self._taxonomy_column.name: self._taxonomy_column.dtype,
            **self.required_context_columns_per_dialect[db_dialect],
        }

    def _get_initial_data(self) -> bach.DataFrame:
        return bach.DataFrame.from_table(
            table_name=self._table_name,
            engine=self._engine,
            index=[],
            all_dtypes=self._get_base_dtypes(),
        )

    def _process_taxonomy_data(self, df: bach.DataFrame) -> bach.DataFrame:
        """
        Extracts all needed values from the taxonomy json/dict and creates a new Series for each.
        Fields are dependent on the engine's taxonomy definition, therefore the new generated series might be
        different per engine.

        Returns a bach DataFrame containing each extracted field as series
        """
        df_cp = df.copy()
        taxonomy_series = df_cp[self._taxonomy_column.name]
        dtypes = _DEFAULT_TAXONOMY_DTYPE_PER_FIELD

        if is_bigquery(df.engine):
            # taxonomy column is a list, we just need to get the first value (LIST IS ALWAYS A SINGLE EVENT)
            taxonomy_series = taxonomy_series.elements[0]

            # same definition as _DEFAULT_TAXONOMY_DTYPE_PER_FIELD with some extra fields
            dtypes = self._taxonomy_column.dtype[0]

        for key, dtype in dtypes.items():
            # parsing element to string and then to dtype will avoid
            # conflicts between casting compatibility
            if isinstance(taxonomy_series, bach.SeriesJson):
                taxonomy_col = taxonomy_series.json.get_value(key, as_str=True)
            else:
                # taxonomy_series is dtype='dict' for BQ. Therefore, we need to explicitly
                # cast the resultant element as string
                taxonomy_col = taxonomy_series.elements[key].astype('string')

            taxonomy_col = taxonomy_col.astype(dtype).copy_override(name=key)
            df_cp[key] = taxonomy_col

        # rename series to objectiv supported
        df_cp = df_cp.rename(
            columns={
                'cookie_id': ObjectivSupportedColumns.USER_ID.value,
                '_type': ObjectivSupportedColumns.EVENT_TYPE.value,
                '_types': ObjectivSupportedColumns.STACK_EVENT_TYPES.value,
            },
        )

        return df_cp

    def _apply_extra_processing(self, df: bach.DataFrame) -> bach.DataFrame:
        """
        Applies more operations to the dataframe (if needed).

        Returns a bach DataFrame
        """
        df_cp = df.copy()
        if is_postgres(self._engine):
            return df_cp

        # remove taxonomy column, no longer needed
        df_cp = df_cp.drop(columns=[self._taxonomy_column.name])

        # this materialization is to generate a readable query
        df_cp = df_cp.materialize(node_name='bq_extra_processing')

        # Remove duplicated event_ids
        # Unfortunately, some events might share an event ID due to
        # browser pre-cachers or scraping bots sending the same event multiple time. Although,
        # legitimate clients might try to send the same events multiple times,
        # in an attempt to make sure that events do not get lost in case of connection problems
        # and/or app reloads. We can easily recognize these events as they'll have non-unique event-ids.
        # In all cases we are only interested in the first event. On postgres we achieve this by having a
        # primary key index on event-id. On BigQuery such indexes are not possible. Instead, we here filter
        # out duplicate event-ids, keeping the first event
        # based on the time the collector sends it to snowplow.
        df_cp = df_cp.drop_duplicates(subset=['event_id'], sort_by=['collector_tstamp'], keep='first')

        # BQ data source has no moment and day columns, therefore we need to generate them
        # based on the time value from the taxonomy column
        df_cp['moment'] = df_cp['time'].copy_override(
            expression=bach.expression.Expression.construct(f'TIMESTAMP_MILLIS({{}})', df_cp['time']),
        )
        df_cp['moment'] = df_cp['moment'].copy_override_type(bach.SeriesTimestamp)
        df_cp['day'] = df_cp['moment'].astype('date')

        return df_cp

    @classmethod
    def validate_pipeline_result(cls, result: bach.DataFrame) -> None:
        """
        Checks if we are returning ALL expected context series with proper dtype.
        """
        check_objectiv_dataframe(
            result,
            columns_to_check=ObjectivSupportedColumns.get_extracted_context_columns(),
            check_dtypes=True,
            infer_identity_resolution=False,
        )

    def _apply_date_filter(
        self,
        df: bach.DataFrame,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        **kwargs,
    ) -> bach.DataFrame:
        """
        Filters dataframe by a date range.

        returns bach DataFrame
        """
        df_cp = df.copy()

        if not start_date and not end_date:
            return df_cp

        date_filters = []
        if start_date:
            date_filters.append(df_cp[self.DATE_FILTER_COLUMN] >= start_date)
        if end_date:
            date_filters.append(df_cp[self.DATE_FILTER_COLUMN] <= end_date)

        return df_cp[reduce(operator.and_, date_filters)]



"""
Copyright 2022 Objectiv B.V.
"""

# Any import from modelhub initializes all the types, do not remove
from modelhub import __version__
from tests_modelhub.data_and_utils.utils import get_objectiv_dataframe_test
from tests.functional.bach.test_data_and_utils import assert_equals_data


def test_top_product_features(db_params):
    df, modelhub = get_objectiv_dataframe_test(db_params)
    initial_columns = df.data_columns
    # without location_stack
    tdf = modelhub.aggregate.top_product_features(df)
    assert len(tdf.index) == 3

    # index _application
    assert_equals_data(
        tdf.index['application'],
        expected_columns=['application'],
        expected_data=[
            ['objectiv-docs'],
            ['objectiv-docs'],
            ['objectiv-docs'],
            ['objectiv-website'],
            ['objectiv-website'],
            ['objectiv-website'],
            ['objectiv-website'],
            ['objectiv-website'],
            ['objectiv-website'],
            ['objectiv-website'],
            ['objectiv-website'],
            ['objectiv-website']
        ],
        order_by=['application'],
        use_to_pandas=True
    )

    # index feature_nice_name
    assert 'feature_nice_name' in tdf.index

    # index event_type
    assert set(tdf.index['event_type'].array) == {'ClickEvent'}

    # data info
    assert list(tdf.data.keys()) == ['user_id_nunique']
    assert set(tdf['user_id_nunique'].array) == {1}

    # with location_stack
    location_stack = df.location_stack.json[{'_type': 'LinkContext'}:]
    tdf = modelhub.aggregate.top_product_features(df, location_stack)
    # make sorting always the same (ignores potential case sensitivity)
    tdf['sort_str'] = tdf.reset_index().feature_nice_name.str[7:]
    assert_equals_data(
        tdf,
        expected_columns=['application', 'feature_nice_name', 'event_type', 'user_id_nunique', 'sort_str'],
        expected_data=[
            ['objectiv-docs', 'Link: logo', 'ClickEvent', 1, 'ogo'],
            ['objectiv-docs', 'Link: notebook-product-analytics', 'ClickEvent', 1, 'otebook-product-analytics'],
            ['objectiv-website', 'Link: About Us', 'ClickEvent', 2, 'bout Us'],
            ['objectiv-website', 'Link: GitHub', 'ClickEvent', 1, 'itHub'],
            ['objectiv-website', 'Link: Docs', 'ClickEvent', 1, 'ocs'],
            ['objectiv-website', 'Link: Contact Us', 'ClickEvent', 1, 'ontact Us'],
            ['objectiv-website', 'Link: Cookies', 'ClickEvent', 1, 'ookies'],
            ['objectiv-website', 'Link: cta-docs-location-stack', 'ClickEvent', 1, 'ta-docs-location-stack'],
            ['objectiv-website', 'Link: cta-docs-taxonomy', 'ClickEvent', 1, 'ta-docs-taxonomy'],
            ['objectiv-website', 'Link: cta-repo-button', 'ClickEvent', 1, 'ta-repo-button']]
,
        use_to_pandas=True,
        order_by=['application', 'sort_str', 'user_id_nunique']

    )

    # check if any new column is added to the original dataframe
    assert sorted(initial_columns) == sorted(df.data_columns)



import inspect
from functools import wraps
from typing import List, Union, Optional, TYPE_CHECKING
from typing_extensions import Protocol

import bach
from sql_models.constants import not_set


if TYPE_CHECKING:
    from modelhub import Map, Aggregate


class ModelFunctionType(Protocol):
    @property
    def __name__(self) -> str:
        ...

    def __call__(
        self, _self: Union['Map', 'Aggregate'], data: bach.DataFrame, *args, **kwargs
    ) -> bach.Series:
        ...


def use_only_required_objectiv_series(
    required_series: Optional[List[str]] = None,
    include_series_from_params: Optional[List[str]] = None,
):
    """
    Internal: Decorator for validating and limiting the series used on a function dedicated
    to generate new aggregated series based on supported objectiv columns.

    :param required_series: A list of objectiv series names that the DataFrame must have
    :param include_series_from_params: A list of parameters containing series names to be considered
    in the dataframe.

    The main purposes of the decorator are to:
        * Validate that the dataframe passed to the function contains all required series for the
        calculation,
        * Optimize the base node of the returned series, as materialization is always applied after
        aggregations. If the passed dataframe includes unused series containing complex expressions,
        such expressions will remain on the final base node's history, meaning that the final query
        might contain extra information that is not related to the expected result. Currently, bach
        does not perform any optimization over this scenario, therefore ModelHub must be in charge
        of ensuring the final query contains only the columns needed for all calculations.
    """
    def check_objectiv_data_decorator(func: ModelFunctionType):
        @wraps(func)
        def wrapped_function(
            _self: Union['Map', 'Aggregate'],
            data: bach.DataFrame,
            *args,
            **kwargs,
        ) -> bach.Series:
            from modelhub.util import check_objectiv_dataframe, ObjectivSupportedColumns
            check_objectiv_dataframe(
                df=data,
                columns_to_check=required_series,
                check_index=True,
                check_dtypes=True,
                with_md_dtypes=True,
            )

            extra_series = _get_extra_series_to_include_from_params(
                func, data, include_series_from_params, *args, **kwargs,
            )
            series_to_include = (
                set(required_series or ObjectivSupportedColumns.get_data_columns()) | set(extra_series)
            )
            data = data[[s for s in data.data_columns if s in series_to_include]]
            return func(_self, data, *args, **kwargs)

        return wrapped_function

    return check_objectiv_data_decorator


def _get_extra_series_to_include_from_params(
    caller: ModelFunctionType,
    data: bach.DataFrame,
    include_series_from_params: Optional[List[str]] = None,
    *args,
    **kwargs
) -> List[str]:
    """
    Helper for use_only_required_objectiv_series decorator. Gets extra series provided via
    the caller's parameters. Will validate that all keyword params exist on the caller's signature
    and that all of them are type string. If a value or default is provided for the parameter, then
    it will be checked if it exists in the objectiv dataframe.

    returns a list of series names
    """
    from modelhub.aggregate import GroupByType
    if not include_series_from_params:
        return []

    extra_series = []
    signature = inspect.signature(caller)
    valid_parameters = list(signature.parameters.keys())

    for keyword_series_incl in include_series_from_params:
        if keyword_series_incl not in valid_parameters:
            raise Exception(f'{keyword_series_incl} does not exist in signature for {caller.__name__}.')

        param = signature.parameters[keyword_series_incl]
        param_index = valid_parameters.index(keyword_series_incl) - 2
        if param.annotation not in (str, GroupByType):
            raise Exception(f'{keyword_series_incl} must be str type.')

        if keyword_series_incl in kwargs:
            series_value = kwargs[keyword_series_incl]
        else:
            series_value = args[param_index] if len(args) > param_index else param.default

        series_to_check = series_value if isinstance(series_value, list) else [series_value]
        for series in series_to_check:
            # ignore if it's None, not_set or a Series
            if series is None or isinstance(series, bach.Series) or series == not_set:
                continue

            if series not in data.data_columns:
                raise ValueError(f'{series} does not exist in objectiv dataframe.')

            extra_series.append(series)
    return extra_series



"""
Copyright 2021 Objectiv B.V.
"""
from typing import TYPE_CHECKING
from bach.series import SeriesInt64

if TYPE_CHECKING:
    from bach.series import SeriesBoolean, Series


class Metrics:
    @staticmethod
    def _cm_prepare(y: 'SeriesBoolean', y_pred: 'SeriesBoolean') -> 'Series':
        df = y.copy_override(name='y').to_frame()
        df['y_pred'] = y_pred
        return df.value_counts()

    @classmethod
    def accuracy_score(cls, y: 'SeriesBoolean', y_pred: 'SeriesBoolean') -> float:
        """
        Returns the accuracy score from a series of true values compared to predicted values.

        :param y: Series with the true target variable.
        :param y_pred: Series with the predicted target variable.
        :returns: a single value with the proportion of correct predicted labels.

        .. note::
            This function queries the database.
        """
        cm = cls._cm_prepare(y, y_pred).to_frame().reset_index()

        if not isinstance(cm['value_counts'], SeriesInt64):
            raise TypeError(f"cm is of type {type(y)}, should be SeriesInt64")

        return (cm[cm.y == cm.y_pred]['value_counts'].sum() / cm['value_counts'].sum()).value
